;;;; utils.scm - XML-RPC support code


(require 'base64)


(declare
  (uses extras lolevel srfi-1 regex)
  (export xml-rpc:marshall-value xml-rpc:unmarshall-value xml-rpc:error) )


(define-constant bad-argument-type-code 5)
(define-constant bad-result-type-code 6)

(define (xml-rpc:marshall-value x)
  (string-append
   "<value>"
   (cond [(number? x)
	  (if (exact? x)
	      (sprintf "<i4>~A</i4>" x)
	      (sprintf "<double>~A</double>" x) ) ]
	 [(boolean? x) (sprintf "<boolean>~A</boolean>" (if x 1 0))]
	 [(string? x)
	  (string-append "<string>" (canonicalize-string x) "</string>") ]
	 [(byte-vector? x) (string-append "<base64>" (base64:encode (byte-vector->string x)) "</base64>")]
	 [(vector? x)
	  (string-append
	   (apply
	    string-append
	    "<array><data>"
	    (map xml-rpc:marshall-value (vector->list x)) )
	   "</data></array>") ]
	 [(and (list? x) (every pair? x))
	  (string-append
	   "<struct>"
	   (fold-right
	    (lambda (x r)
	      (string-append
	       "<member><name>" (canonicalize-string (->string (car x))) "</name>"
	       (xml-rpc:marshall-value (cdr x))
	       "</member>"
	       r) )
	    "</struct>"
	    x) ) ]
	 [(list? x)
	  (string-append
	   (apply
	    string-append
	    "<array><data>"
	    (map xml-rpc:marshall-value x) )
	   "</data></array>") ]
	 ;; missing: dateTime.iso8601
	 [else (xml-rpc:error bad-argument-type-code "argument type not supported" x)] )
   "</value>") )

(define xml-rpc:unmarshall-value
  (match-lambda
    [('value x)
     (match x
       [((or 'i4 'int 'double) n) (string->number n)]
       [('boolean b) (string=? "1" b)]
       [('string) ""]
       [('string s) s]
       [('base64 s) (string->byte-vector (base64:decode s))]
       [`(struct (member (name ,names) ,vals) ...)
	(map cons names (map xml-rpc:unmarshall-value vals)) ]
       [`(array (data . ,vals))
	(list->vector (map xml-rpc:unmarshall-value vals)) ]
       [`(|dateTime.iso8601| ,dt) dt]
       [_ x] ) ]
    [result (xml-rpc:error bad-result-type-code "invalid response format: bad result" result)] ) )

(define (canonicalize-string str)
  ;; this could be done better...
  (string-substitute "<" "&lt;" (string-substitute "&" "&amp;" str #t) #t) )

(define (xml-rpc:error code msg . args)
  (signal
   (make-composite-condition
    (make-property-condition 'exn 'message msg 'arguments args)
    (make-property-condition 'xml-rpc 'code code) ) ) )
